/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.router;

import java.util.Collection;
import java.util.Collections;
import java.util.Optional;

import jakarta.annotation.PostConstruct;
import jakarta.validation.Valid;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import de.ozgcloud.eingang.router.VorgangManagerListProperties.FallbackStrategy;
import de.ozgcloud.eingang.router.VorgangManagerListProperties.RoutingStrategy;
import de.ozgcloud.eingang.router.errorhandling.AdapterConfigurationException;
import de.ozgcloud.eingang.router.errorhandling.UnknownOrganisationseinheitException;
import de.ozgcloud.vorgang.grpc.binaryFile.BinaryFileServiceGrpc.BinaryFileServiceStub;
import de.ozgcloud.vorgang.vorgang.VorgangServiceGrpc.VorgangServiceBlockingStub;
import io.grpc.Channel;
import io.grpc.stub.AbstractStub;
import lombok.NonNull;
import lombok.extern.log4j.Log4j2;
import net.devh.boot.grpc.client.channelfactory.GrpcChannelFactory;
import net.devh.boot.grpc.client.inject.StubTransformer;
import net.devh.boot.grpc.client.stubfactory.StubFactory;

@Component
@Log4j2
public class VorgangManagerServerResolver {

	static final String CHANNEL_NAME_PREFIX = "vorgang-manager-";

	@Autowired
	private GrpcChannelFactory grpcChannelFactory;
	@Autowired
	@Valid
	private VorgangManagerListProperties properties;

	@Autowired(required = false)
	private Collection<StubFactory> stubFactories = Collections.emptyList();
	@Autowired(required = false)
	private Collection<StubTransformer> stubTransformers = Collections.emptyList();

	private StubFactory vorgangStubFactory;
	private StubFactory binaryFileAsynStubFactory;

	@PostConstruct
	void findApplicableStubFactories() {
		vorgangStubFactory = findStubFactory(VorgangServiceBlockingStub.class);
		binaryFileAsynStubFactory = findStubFactory(BinaryFileServiceStub.class);
	}

	StubFactory findStubFactory(Class<? extends AbstractStub<?>> stubClass) {
		return stubFactories.stream()
				.filter(factory -> factory.isApplicable(stubClass))
				.findFirst().orElseThrow(() -> new AdapterConfigurationException("Cannot find Stub-Factory for GRPC-" + stubClass));
	}

	public VorgangServiceBlockingStub resolveVorgangServiceBlockingStubByOrganisationseinheitenId(Optional<String> organisationsEinheitId) {
		return (VorgangServiceBlockingStub) createStub(organisationsEinheitId, vorgangStubFactory, VorgangServiceBlockingStub.class);
	}

	public BinaryFileServiceStub resolveBinaryFileServiceStubByOrganisationsEinheitId(Optional<String> organisationsEinheitId) {
		return (BinaryFileServiceStub) createStub(organisationsEinheitId, binaryFileAsynStubFactory, BinaryFileServiceStub.class);
	}

	AbstractStub<?> createStub(Optional<String> organisationsEinheitId, StubFactory stubFactory, Class<? extends AbstractStub<?>> stubClass) {// NOSONAR
		var channelName = getChannelName(organisationsEinheitId);
		var stub = stubFactory.createStub(stubClass, createChannelByName(channelName));

		return applyStubTransformers(stub, channelName);
	}

	AbstractStub<?> applyStubTransformers(AbstractStub<?> stub, String channelName) { // NOSONAR wildcard given by StubTransformer
		for (var transformer : stubTransformers) {
			stub = transformer.transform(channelName, stub);
		}
		return stub;
	}

	String getChannelName(Optional<String> organisationsEinheitId) {
		Optional<String> target;

		if (properties.getRoutingStrategy() == RoutingStrategy.SINGLE) {
			target = properties.getTargetVorgangManagerName();
		} else {
			target = organisationsEinheitId.map(properties.getOrganisationseinheiten()::get);
		}

		return target.map(this::addChannelPrefix).orElseGet(this::getFundstelleChannelName);
	}

	private String addChannelPrefix(@NonNull String name) {
		return CHANNEL_NAME_PREFIX + name;
	}

	String getFundstelleChannelName() {
		return properties.getFallbackStrategy().map(fallbackStrategy -> {
			if (fallbackStrategy == FallbackStrategy.DENY) {
				throw new UnknownOrganisationseinheitException();
			} else {
				return properties.getFundstelleVorgangManagerName().map(this::addChannelPrefix).orElseThrow(() -> new AdapterConfigurationException(
						"Property 'fundstelleVorgangManagerName' is missing but required for fallbackStrategy 'FUNDSTELLE'"));
			}
		}).orElseThrow(() -> {
			LOG.warn("Missing required routing fallback Strategy. Falling back to 'DENY'");
			return new UnknownOrganisationseinheitException();
		});
	}

	Channel createChannelByName(String name) {
		return grpcChannelFactory.createChannel(name);
	}
}