package de.ozgcloud.eingang.xta;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.eingang.common.formdata.IncomingFileTestFactory;
import de.ozgcloud.eingang.common.vorgang.VorgangNummerSupplier;
import de.ozgcloud.eingang.xdomea.XdomeaMessageData;
import de.ozgcloud.eingang.xdomea.XdomeaMessageDataMapper;
import de.ozgcloud.eingang.xdomea.XdomeaMessageDataTestFactory;

class XtaServiceTest {

	private static final String MESSAGE_TYPE_OTHER = "MESSAGE_TYPE_OTHER";

	@Spy
	@InjectMocks
	private XtaService service;

	@Mock
	private XtaRemoteService remoteService;
	@Mock
	private XtaMessageMapper mapper;
	@Mock
	private VorgangNummerSupplier vorgangNummerSupplier;

	@Mock
	private XtaIncomingFilesMapper incomingFilesMapper;

	@Mock
	private XdomeaMessageDataMapper xdomeaMessageDataMapper;

	@Mock
	private XtaCurrentIdentifierService currentIdentifierService;

	@Nested
	class TestGetMessagesAsFormData {

		private final XtaMessageMetaData messageMetaData = XtaMessageMetaDataTestFactory.create();
		private final XtaMessage message = XtaMessageTestFactory.create();

		@BeforeEach
		void setup() {
			doReturn(Stream.of(messageMetaData)).when(service).createXtaMessageStream();
		}

		@Test
		void shouldCallCreateStream() {
			service.getMessages();

			verify(service).createXtaMessageStream();
		}

		@Test
		void shouldCallFilterByMessageType() {
			setupMocks();

			service.getMessages().toList();

			verify(service).isSupportedMessageType(messageMetaData);
		}

		@Test
		void shouldCallGetFormData() {
			setupMocks();
			doReturn(true).when(service).isSupportedMessageType(messageMetaData);

			service.getMessages().toList();

			verify(service).getFormData(messageMetaData);
		}

		@Test
		void shouldNotCallGetFormData() {
			doReturn(false).when(service).isSupportedMessageType(messageMetaData);

			service.getMessages().toList();

			verify(service, never()).getFormData(any());
		}

		private void setupMocks() {
			var testFormData = FormDataTestFactory.create();
			when(mapper.toFormData(any(), any(), eq(vorgangNummerSupplier))).thenReturn(testFormData);
			when(remoteService.getMessage(any(XtaMessageId.class))).thenReturn(message);
		}
	}

	@Nested
	class TestIsSupportedMessageType {

		@Test
		void shouldAcceptDFoerdermittel() {
			var metaDataDFoerder = XtaMessageMetaDataTestFactory.create();

			assertThat(service.isSupportedMessageType(metaDataDFoerder)).isTrue();
		}

		@Test
		void shouldNotAcceptOtherMessageType() {
			var metaDataDFoerder = XtaMessageMetaDataTestFactory.createBuilder().messageType(MESSAGE_TYPE_OTHER).build();

			assertThat(service.isSupportedMessageType(metaDataDFoerder)).isFalse();
		}
	}

	@Nested
	class TestGetFormData {

		private XtaMessage message;
		private XdomeaMessageData classification;

		@Captor
		private ArgumentCaptor<XtaMessageMetaData> messageMetaDataCaptor;

		@Captor
		private ArgumentCaptor<XdomeaMessageData> classificationCaptor;

		@BeforeEach
		void init() {
			message = XtaMessageTestFactory.create();
			when(remoteService.getMessage(any(XtaMessageId.class))).thenReturn(message);

			classification = XdomeaMessageDataTestFactory.create();
			var incomingFiles = List.of(IncomingFileTestFactory.create(), IncomingFileTestFactory.create());
			when(incomingFilesMapper.toIncomingFiles(message.getMessageFiles())).thenReturn(incomingFiles);
			classification = XdomeaMessageDataTestFactory.create();
			when(xdomeaMessageDataMapper.mapIncomingFilesToXdomeaMessageData(incomingFiles)).thenReturn(classification);
		}

		@Test
		void shouldCallRemoteService() {
			service.getFormData(XtaMessageMetaDataTestFactory.create());

			verify(remoteService).getMessage(XtaMessageTestFactory.MESSAGE_ID);
		}

		@Test
		void shouldCallMapper() {
			service.getFormData(XtaMessageMetaDataTestFactory.create());

			verify(mapper).toFormData(any(), any(), eq(vorgangNummerSupplier));
		}

		@Test
		void shouldHaveMetaData() {
			XtaMessageMetaData metaData = XtaMessageMetaDataTestFactory.create();

			service.getFormData(metaData);

			verify(mapper).toFormData(classificationCaptor.capture(), messageMetaDataCaptor.capture(), eq(vorgangNummerSupplier));
			assertThat(messageMetaDataCaptor.getValue()).isEqualTo(metaData);
			assertThat(classificationCaptor.getValue()).isEqualTo(classification);
		}

		@Test
		void shouldReturnMappedResult() {
			var mapped = FormDataTestFactory.create();
			when(mapper.toFormData(any(), any(), eq(vorgangNummerSupplier))).thenReturn(mapped);

			var result = service.getFormData(XtaMessageMetaDataTestFactory.create());

			assertThat(result).isSameAs(mapped);
		}
	}

	@Nested
	class TestAcknowledgeReceive {

		@Test
		void shouldCallRemoteService() {
			service.acknowledgeReceive(XtaMessageTestFactory.MESSAGE_ID);

			verify(remoteService).close(XtaMessageTestFactory.MESSAGE_ID);
		}
	}

	@DisplayName("create XTA message stream")
	@Nested
	class TestCreateXtaMessageStream {
		private static final List<String> XTA_IDENTIFIERS = List.of("XTA_IDENTIFIER", "XTA_IDENTIFIER2");

		@Captor
		private ArgumentCaptor<String> identifierCaptor;

		@BeforeEach
		void mock() {
			when(remoteService.getMessagesMetadata())
					.thenReturn(XtaMessageMetaDatasAndHeaderTestFactory.create())
					.thenReturn(XtaMessageMetaDatasAndHeaderTestFactory.create());
			when(currentIdentifierService.getIdentifiers()).thenReturn(XTA_IDENTIFIERS);
		}

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			var result = service.createXtaMessageStream().toList();

			assertThat(result).isEqualTo(getExpectedMessages());
		}

		private List<XtaMessageMetaData> getExpectedMessages() {
			return Stream.concat(
					XtaMessageMetaDatasAndHeaderTestFactory.create().getMessages(),
					XtaMessageMetaDatasAndHeaderTestFactory.create().getMessages()
			).toList();
		}

		@DisplayName("should call set current identifier")
		@Test
		void shouldCallSetCurrentIdentifier() {
			service.createXtaMessageStream().toList();

			verify(currentIdentifierService, times(2))
					.setCurrentIdentifier(identifierCaptor.capture());
			assertThat(identifierCaptor.getAllValues()).isEqualTo(XTA_IDENTIFIERS);
		}
	}

}
